# -*- coding: utf-8 -*-

import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


def _init_builder(type_):
    def __init__(self, *args, **kwargs):
        super(type(self), self).__init__()
        self.encoder = type_(*args, **kwargs)

    return __init__


def forward(
    self,
    inputs,
    lengths,
    batch_first=True,
    enforce_sorted=False,
    padding_value=0.0,
    total_length=None,
):
    # ATM `lengths` must be 1d long tensor on CPU.
    # https://github.com/pytorch/pytorch/issues/43227
    packed = pack_padded_sequence(
        inputs, lengths.cpu(), batch_first=batch_first, enforce_sorted=enforce_sorted
    )
    hidden, states = self.encoder(packed)
    hidden, _ = pad_packed_sequence(
        hidden,
        batch_first=batch_first,
        padding_value=padding_value,
        total_length=total_length,
    )

    return hidden, states


_recurrent_types = {
    "RNN": nn.RNN,
    "GRU": nn.GRU,
    "LSTM": nn.LSTM,
}


def _init():
    for name, type_ in _recurrent_types.items():
        class_ = type(
            name, (nn.Module,), {"__init__": _init_builder(type_), "forward": forward}
        )
        globals()[name] = class_


_init()


def get_rnn(cell: str) -> nn.Module:
    return globals()[cell.upper()]



# -*- coding: utf-8 -*-

from __future__ import annotations

from collections.abc import Callable, Mapping

import torch


def get_pooling(method: str) -> Callable[[Mapping, torch.Tensor], torch.Tensor]:
    methods = {
        "cls": cls_pooling,
        "mean": mean_pooling,
        "max": max_pooling,
    }
    pooling = methods.get(method)
    if pooling is None:
        raise ValueError(f"Pooling `method` should be one of {list(methods)}")

    return pooling


def mean_pooling(model_output: Mapping, attention_mask: torch.Tensor) -> torch.Tensor:
    token_embeddings = model_output["last_hidden_state"]
    input_mask_expanded = attention_mask.unsqueeze(dim=-1).float()
    sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, dim=1)
    sum_mask = torch.clamp(input_mask_expanded.sum(dim=1), min=1e-9)

    return sum_embeddings / sum_mask


def max_pooling(model_output: Mapping, attention_mask: torch.Tensor) -> torch.Tensor:
    token_embeddings = model_output["last_hidden_state"]
    input_mask_expanded = (
        attention_mask.unsqueeze(dim=-1).expand(token_embeddings.size()).float()
    )
    token_embeddings[input_mask_expanded == 0] = -1e9
    max_over_time = torch.max(token_embeddings, dim=1)[0]

    return max_over_time


def cls_pooling(model_output: Mapping, attention_mask: torch.Tensor) -> torch.Tensor:
    return model_output["last_hidden_state"][:, 0]



# -*- coding: utf-8 -*-

import torchlight.attention
import torchlight.dataset
import torchlight.logger
import torchlight.losses
import torchlight.masking
import torchlight.optim
import torchlight.pooling
import torchlight.preprocessing
import torchlight.rnn
import torchlight.scheduled_sampling
import torchlight.training
import torchlight.utils

__version__ = "0.1.1"



# -*- coding: utf-8 -*-

from __future__ import annotations

import itertools
import json
import os
from collections.abc import Iterable
from typing import Optional, Type, TypeVar, Union

import torch


class LabelEncoder(object):
    T = TypeVar("T", bound="LabelEncoder")

    def __init__(
        self, labels: Optional[Iterable[str]] = None, default: Optional[str] = None
    ) -> None:
        self.default = default
        self.init(labels)

    def __len__(self):
        return len(self.index2label)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(labels={self.labels!r}, default={self.default!r})"
        )

    @property
    def labels(self):
        return list(self.label2index)

    @property
    def num_labels(self):
        return len(self)

    @property
    def default_index(self):
        return self.get_index(self.default)

    def reset(self) -> None:
        self.index2label = {}
        self.label2index = {}

        if self.default:
            self.label2index[self.default] = 0
            self.index2label[0] = self.default

    def init(self, labels: Optional[Iterable[str]]) -> None:
        self.reset()

        if not labels:
            return

        for label in labels:
            self.label2index.setdefault(label, len(self.label2index))

        self.index2label = {v: k for k, v in self.label2index.items()}

    def add(self, label: str) -> int:
        index = self.label2index.setdefault(label, len(self))
        self.index2label.setdefault(index, label)

        return index

    def get_index(self, label: str) -> int:
        index = self.label2index.get(label)
        if index is None:
            if self.default is not None:
                return 0

            raise KeyError(label)

        return index

    def get_label(self, index: int) -> str:
        label = self.index2label.get(index)
        if label is None:
            if self.default is not None:
                return self.default

            raise KeyError(index)

        return label

    def encode_label(
        self, label: str, return_tensors: Optional[str] = None
    ) -> Union[int, torch.Tensor]:
        index = self.get_index(label)

        if return_tensors == "pt":
            return torch.tensor(index, dtype=torch.int64)

        if isinstance(return_tensors, str):
            raise ValueError('`return_tensors` should be "pt" or None')

        return index

    def decode_label(self, index: Union[int, torch.LongTensor]) -> str:
        if isinstance(index, torch.LongTensor):
            if index.ndim > 0:
                raise ValueError(
                    f"tensor should be 0-d tensor, got: ndim == {index.ndim}"
                )
            index = int(index.cpu().numpy())

        if not isinstance(index, int):
            raise ValueError(
                "`index` should be int or torch.LongTensor, "
                f"got: {index.__class__.__name__}"
            )

        return self.get_label(index)

    def encode(
        self, labels: Iterable[str], return_tensors: Optional[str] = None
    ) -> Union[list[int], torch.Tensor]:
        indices = [self.get_index(x) for x in labels]
        if return_tensors == "pt":
            return torch.tensor(indices, dtype=torch.int64)

        if isinstance(return_tensors, str):
            raise ValueError('`return_tensors` should be "pt" or None')

        return indices

    def decode(self, indices: Union[torch.LongTensor, Iterable[int]]) -> list[str]:
        if isinstance(indices, torch.LongTensor):
            if indices.ndim != 1:
                raise ValueError(
                    f"tensor should be 1-d tensor, got: ndim == {indices.ndim}"
                )
            indices = indices.cpu().numpy()

        labels = [self.get_label(x) for x in indices]

        return labels

    def save(self, filename: Union[str, os.PathLike]) -> None:
        with open(filename, mode="w") as f:
            data = {
                "labels": self.labels,
                "default": self.default,
            }
            json.dump(data, f, ensure_ascii=False, indent=4)

    @classmethod
    def from_iterable(cls: Type[T], labels: Iterable[Iterable[str]]) -> T:
        return cls(itertools.chain.from_iterable(labels))

    @classmethod
    def load(cls: Type[T], filename: Union[str, os.PathLike]) -> T:
        with open(filename) as f:
            data = json.load(f)

            return cls(data["labels"], default=data["default"])



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Optional, Union

import torch


def create_span_mask(
    starts: Union[list[int], torch.Tensor],
    ends: Union[list[int], torch.Tensor],
    length: int,
    dtype: torch.dtype = torch.int64,
    device: torch.device = torch.device("cpu"),
) -> torch.Tensor:
    if len(starts) != len(ends):
        raise ValueError(
            f"`start` and `end` should have same lengths: {len(starts)} != {len(ends)}"
        )

    if len(starts) == 0:
        return torch.zeros((0, length), dtype=dtype, device=device)

    def _convert_tensor(t, name):
        if isinstance(t, torch.Tensor):
            if t.ndim != 1:
                raise ValueError(
                    f"`{name}` must be 1d if passed as tensor, got ndim == {t.ndim}"
                )

            return t.clone().detach()

        return torch.tensor(t, dtype=dtype, device=device)

    start, end = _convert_tensor(starts, "starts"), _convert_tensor(ends, "ends")

    mask = torch.arange(length, dtype=dtype, device=device)
    mask = (start[:, None] <= mask) & (mask < end[:, None])
    mask = mask.type_as(start)

    return mask


def length_to_mask(
    length: torch.Tensor,
    max_len: Optional[Union[int, torch.Tensor]] = None,
    batch_first: bool = False,
    dtype: torch.dtype = torch.int64,
    device: Optional[torch.device] = None,
) -> torch.Tensor:
    if max_len is None:
        max_len = length.max()

    if device is None:
        device = length.device

    mask = torch.arange(max_len, device=device).unsqueeze(dim=1).expand(
        max_len, len(length)
    ) < length.unsqueeze(dim=0)
    mask = mask.type(dtype)

    return mask.transpose(0, 1) if batch_first else mask


def mask_to_length(
    mask: torch.Tensor,
    batch_first: bool = False,
    dtype: torch.dtype = torch.int64,
    device: Optional[torch.device] = None,
) -> torch.Tensor:
    length = mask.sum(dim=int(batch_first)).type(dtype)
    if device is not None:
        length = length.to(device)

    return length



# -*- coding: utf-8 -*-

from typing import Union

import torch
import torch.nn as nn


def manhattan_similarity(v1: torch.Tensor, v2: torch.Tensor) -> torch.Tensor:
    return torch.exp(-torch.norm(v1 - v2, p=1, dim=-1, keepdim=True))


class ManhattanSimilarity(nn.Module):
    def forward(
        self, v1: torch.Tensor, v2: torch.Tensor
    ) -> torch.Tensor:  # pylint: disable=no-self-use
        return manhattan_similarity(v1, v2)


class HingeLoss(nn.Module):
    def __init__(self, margin: Union[float, torch.Tensor] = 1.0):
        super().__init__()
        self.margin = margin

    def forward(
        self, positive_logits: torch.Tensor, negative_logits: torch.Tensor
    ) -> torch.Tensor:
        return torch.mean(
            torch.clamp(negative_logits + self.margin - positive_logits, 0)
        )



# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import itertools
import warnings
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Optional, TypeVar, Union, cast

import ignite.distributed as idist
import torch
from ignite.distributed.auto import DistributedProxySampler
from torch.utils.data import BatchSampler, DataLoader, Dataset
from torch.utils.data import IterableDataset as _IterableDataset
from torch.utils.data.dataset import IterableDataset
from torch.utils.data.sampler import RandomSampler, Sampler, SequentialSampler

from torchlight.utils import ModeKeys, PhaseMixin


class Collator(PhaseMixin, metaclass=abc.ABCMeta):
    T = TypeVar("T", bound="Collator")

    def __init__(self, mode: ModeKeys = ModeKeys.TRAIN) -> None:
        self.mode = mode

    def __call__(self, batch: Sequence) -> Any:
        return self.collate_fn(batch)

    def encode(self, example) -> Any:  # pylint: disable=no-self-use
        return example

    def encode_batch(self, batch: Sequence) -> list:
        return list(map(self.encode, batch))

    def collate_train(self, batch: Sequence) -> Any:
        raise NotImplementedError()

    def collate_eval(self, batch: Sequence) -> Any:
        return self.collate_train(batch)

    def collate_predict(self, batch: Sequence) -> Any:
        return self.collate_train(batch)

    def _get_collate_fn(self):
        return {
            ModeKeys.TRAIN: self.collate_train,
            ModeKeys.EVAL: self.collate_eval,
            ModeKeys.PREDICT: self.collate_predict,
        }[self.mode]

    def _collate(self, encoded_batch):
        return self._get_collate_fn()(encoded_batch)

    def collate_fn(self, batch: Sequence) -> Any:
        encoded_batch = self.encode_batch(batch)
        collated = self._collate(encoded_batch)

        return collated


def _identity(x):
    return x


def _random_access_bucket(
    bucket: list, batch_size: int, drop_last: Optional[bool] = None
) -> Iterable:
    while bucket:
        size = len(bucket)
        if size <= batch_size:
            if size == batch_size or drop_last is False:
                yield bucket
            bucket = []
        else:
            # NOTE: Use `torch.randint` becasue
            # `DistributedProxySampler` use `torch.manual_seed`.
            start = cast(int, torch.randint(0, size - batch_size, size=(1,)).item())
            end = start + batch_size
            yield bucket[start:end]
            bucket = bucket[:start] + bucket[end:]


class BucketBatchSampler(BatchSampler):
    def __init__(
        self,
        dataset: Dataset,
        sampler: Sampler,
        batch_size: int,
        drop_last: bool,
        sort_key: Callable = _identity,
        batch_size_multiplier: int = 100,
    ) -> None:
        super().__init__(sampler, batch_size, drop_last)

        self.dataset = dataset
        self.bucket_size = min(
            batch_size_multiplier * batch_size, len(sampler)  # type: ignore
        )
        self.batch_sampler = BatchSampler(sampler, self.bucket_size, False)
        self.sort_key = sort_key

    def __iter__(self):
        for bucket in self.batch_sampler:
            bucket.sort(key=lambda i: self.sort_key(self.dataset[i]))

            yield from _random_access_bucket(bucket, self.batch_size, self.drop_last)


def bucket_batch_sampler(
    dataset: Dataset,
    batch_size: int,
    drop_last: bool,
    sort_key: Callable = _identity,
    batch_size_multiplier: int = 100,
) -> DistributedProxySampler:
    return DistributedProxySampler(
        BucketBatchSampler(
            dataset,
            RandomSampler(range(len(dataset))),  # type: ignore
            batch_size,
            drop_last,
            sort_key=sort_key,
            batch_size_multiplier=batch_size_multiplier,
        ),
        num_replicas=idist.get_world_size(),
        rank=idist.get_rank(),
    )


class IterableDataset(_IterableDataset):
    def __init__(
        self,
        generating_function: Callable,
        batch_size: int,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        mode: ModeKeys = ModeKeys.TRAIN,  # TODO: Make wrapper.
    ) -> None:
        super().__init__()
        self.generating_function = generating_function
        self.batch_size = batch_size
        if ((rank is None) + (world_size is None)) % 2 != 0:
            raise ValueError("`rank` and `world_size` must both given or unspecified")
        self.world_size = world_size
        self.rank = rank
        self.mode = mode

    def _get_iterator(self):
        # Generate new iterator.
        iterator = self.generating_function()

        # Deal with `DistributedDataParallel`.
        if self.world_size is not None and self.rank is not None:
            iterator = itertools.islice(iterator, self.rank, None, self.world_size)

        # Deal with `num_workers` > 1 of `DataLoader`.
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            iterator = itertools.islice(
                iterator, worker_info.id, None, worker_info.num_workers
            )

        return iterator

    def __getitem__(self, index):
        raise RuntimeError("`IterableDataset` does not support indexing")

    def __iter__(self):
        yield from self._get_iterator()


class BucketIterableDataset(IterableDataset):
    def __init__(
        self,
        generating_function: Callable,
        batch_size: int,
        world_size: Optional[int] = None,
        rank: Optional[int] = None,
        sort_key: Callable = _identity,
        batch_size_multiplier: int = 100,
        mode: ModeKeys = ModeKeys.TRAIN,  # TODO: Make wrapper.
    ) -> None:
        super().__init__(
            generating_function, batch_size, world_size=world_size, rank=rank, mode=mode
        )
        self.bucket_size = batch_size_multiplier * batch_size
        self.sort_key = sort_key

    def __iter__(self):
        iterator = self._get_iterator()

        while bucket := list(itertools.islice(iterator, None, self.bucket_size)):
            bucket.sort(key=self.sort_key)

            for batch in _random_access_bucket(bucket, self.batch_size, False):
                yield from batch


def get_sampler(
    dataset: Dataset, train: bool
) -> Union[RandomSampler, SequentialSampler]:
    if train:
        sampler = RandomSampler(dataset)  # type: ignore
    else:
        sampler = SequentialSampler(dataset)  # type: ignore

    return sampler


def get_dataloader(
    dataset: Dataset,
    collate_fn: Optional[Union[Callable, Collator]] = None,
    sampler: Optional[Sampler] = None,
    batch_sampler: Optional[Sampler] = None,
    sort_key: Optional[Callable] = None,
    shuffle: bool = False,
    **kwargs,
) -> DataLoader:
    if not isinstance(dataset, IterableDataset):
        if batch_sampler is None:
            if sort_key is not None:
                if isinstance(collate_fn, Collator) and collate_fn.is_train():
                    warnings.warn(
                        "`sort_key` is given when `collate_fn.is_train()` is False"
                    )

                batch_sampler = bucket_batch_sampler(
                    dataset,  # type: ignore
                    kwargs["batch_size"],
                    drop_last=False,
                    sort_key=sort_key,
                )

                # When `batch_sampler` is given, `batch_size` must be 1
                # when initializing `DataLoader`.
                kwargs["batch_size"] = 1
                kwargs["batch_sampler"] = batch_sampler

            elif sampler is None:
                # `sampler` will be wrapped in `idist.auto_dataloader`,
                # so we dont' need `batch_sampler`.
                sampler = get_sampler(
                    dataset,
                    collate_fn.is_train()
                    if isinstance(collate_fn, Collator)
                    else shuffle,
                )

    kwargs.update(
        {
            "collate_fn": collate_fn,
            "sampler": sampler,
        }
    )

    dataloader = idist.auto_dataloader(dataset, **kwargs)  # type: DataLoader[Dataset]

    return dataloader



# -*- coding: utf-8 -*-

from __future__ import annotations

import enum
import inspect
import os
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from ignite.handlers.checkpoint import Checkpoint
from transformers.optimization import AdamW, get_linear_schedule_with_warmup


class ModeKeys(str, enum.Enum):
    TRAIN = "train"
    EVAL = "eval"
    PREDICT = "predict"


class PhaseMixin(object):
    mode = ModeKeys.TRAIN

    def train(self) -> None:
        self.mode = ModeKeys.TRAIN

    def eval(self) -> None:
        self.mode = ModeKeys.EVAL

    def predict(self) -> None:
        self.mode = ModeKeys.PREDICT

    def is_train(self) -> bool:
        return self.mode == ModeKeys.TRAIN


def cuda(enable: bool) -> bool:
    return enable and torch.cuda.is_available()


def device(
    device_str: str = "", return_string: bool = True
) -> Union[str, torch.device]:
    if (
        not device_str
        or device_str.startswith("cuda")
        and not torch.cuda.is_available()
    ):
        device_str = "cpu"

    return device_str if return_string else torch.device(device_str)


def split_tensors(
    tensors: Iterable[torch.Tensor], chunk_size: int, dim: int = 0
) -> list[tuple[torch.Tensor, ...]]:
    sizes = [t.size() for t in tensors]
    if not sizes:
        raise ValueError("At least one tensor must be given.")

    for prev, this in zip(sizes, sizes[1:]):
        if prev[dim] != this[dim]:
            raise ValueError(
                f"Size of dimension `dim` = {dim}"  # type: ignore
                " must be same for all input tensors, "
                f"got: {list(map(tuple, sizes))}."
            )

    splits = list(zip(*(torch.split(x, chunk_size, dim=dim) for x in tensors)))

    return splits


def split_apply(
    f: Callable,
    inputs: Iterable[torch.Tensor],
    chunk_size: int,
    *args,
    dim: int = 0,
    **kwargs,
) -> Union[torch.Tensor, tuple[torch.Tensor, ...]]:
    chunks = split_tensors(inputs, chunk_size=chunk_size, dim=dim)
    outputs = [f(*chunk, *args, **kwargs) for chunk in chunks]

    if all(isinstance(t, torch.Tensor) for t in outputs):
        return torch.cat(outputs, dim=dim)

    outputs = list(zip(*outputs))
    tuple_outputs = tuple(torch.cat(x, dim=dim) for x in outputs)

    return tuple_outputs


def check_tensors_dimension(tensors: Iterable[torch.Tensor], dim: int):
    if any(not isinstance(x, torch.Tensor) or x.ndim != dim for x in tensors):
        raise ValueError(f"`tensors` must be {dim}d tensors.")


def pad_stack_1d(
    tensors: Sequence[torch.Tensor],
    max_length: Optional[int] = None,
    return_lengths: bool = False,
) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    # NOTE: Works for 1d tensors with size 0.

    check_tensors_dimension(tensors, 1)

    if max_length is None:
        length = torch.tensor([len(x) for x in tensors], dtype=torch.int64)
        max_length = max(length)
    else:
        if return_lengths:
            raise ValueError(
                "`return_lengths` must be False when `max_length` is given"
            )

    tensor = torch.stack([F.pad(x, [0, max_length - len(x)]) for x in tensors])

    if return_lengths:
        return tensor, length

    return tensor


def pad_stack_2d(
    tensors: Sequence[torch.Tensor], max_rows: int, max_columns: int
) -> torch.Tensor:
    # NOTE: Works for 2d tensors with size 0.

    check_tensors_dimension(tensors, 2)

    # https://discuss.pytorch.org/t/padding-zero-size-tensors/118777
    if max_rows == 0:
        return tensors[0].new_zeros(len(tensors), max_rows, max_columns)

    return torch.stack(
        [
            torch.nn.functional.pad(
                x, [0, max_columns - x.size(1), 0, max_rows - x.size(0)]
            )
            for x in tensors
        ]
    )


def get_default_arguments(f: Callable) -> dict:
    return {
        key: value.default
        for key, value in inspect.signature(f).parameters.items()
        if value.default is not value.empty
    }


def load_checkpoint(
    to_load: Union[Mapping, nn.Module], checkpoint: Union[str, os.PathLike]
) -> None:
    if isinstance(to_load, nn.Module):
        to_load = {"model": to_load}

    ckpt = torch.load(checkpoint, map_location="cpu")
    Checkpoint.load_objects(to_load=to_load, checkpoint=ckpt)


def plm_path(path: Union[str, os.PathLike]) -> str:
    root = os.getenv("PRETRAINED_MODEL_DIR")
    if root:
        return os.path.join(root, path)

    return str(path)


def get_pretrained_optimizer_and_scheduler(
    model: nn.Module,
    lr: float,
    weight_decay: float,
    warmup_steps: int,
    num_training_steps: int,
    optimizer_kwargs: Optional[dict] = None,
    scheduler_kwargs: Optional[dict] = None,
) -> tuple[optim.Optimizer, optim.lr_scheduler.LambdaLR]:
    if lr <= 0:
        raise ValueError("`lr` must be positive.")

    if warmup_steps <= 0:
        raise ValueError("`warmup_steps` must be positive")

    if not optimizer_kwargs:
        optimizer_kwargs = {}

    if not scheduler_kwargs:
        scheduler_kwargs = {}

    no_decay = ["bias", "LayerNorm.weight"]
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if not any(nd in n for nd in no_decay)
            ],
            "weight_decay": weight_decay,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if any(nd in n for nd in no_decay)
            ],
            "weight_decay": 0.0,
        },
    ]
    optimizer = AdamW(optimizer_grouped_parameters, lr=lr, **optimizer_kwargs)

    lr_scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=warmup_steps,
        num_training_steps=num_training_steps,
        **scheduler_kwargs,
    )

    return (optimizer, lr_scheduler)



# -*- coding: utf-8 -*-

from __future__ import annotations

import logging
import os
from collections.abc import Callable, Mapping
from typing import Any, Optional, Union, cast

import ignite.distributed as idist
import torch
import torch.nn as nn
from ignite.contrib.engines.common import (
    add_early_stopping_by_val_score,
    save_best_model_by_val_score,
    setup_wandb_logging,
)
from ignite.contrib.handlers import ProgressBar
from ignite.contrib.handlers.base_logger import BaseLogger
from ignite.engine import Engine, Events
from ignite.engine.events import CallableEventWithFilter
from ignite.handlers import EarlyStopping
from ignite.handlers.checkpoint import Checkpoint
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data.dataloader import DataLoader

from torchlight.dataset import Collator
from torchlight.logger import setup_tb_logging
from torchlight.training.params import Params
from torchlight.utils import ModeKeys

logger = logging.getLogger(__name__)


def get_event(steps: Union[int, str]) -> CallableEventWithFilter:
    if isinstance(steps, int):
        return Events.ITERATION_COMPLETED(every=steps)

    if steps == "epoch":
        return Events.EPOCH_COMPLETED(every=1)

    raise ValueError(
        (
            "Event frequency should be either"
            " integer for Events.ITERATION_COMPLETED"
            ' or "epoch" for Events.EPOCH_COMPLETED'
        )
    )


def handle_collator_mode(engine: Engine, dataset: str, mode: str) -> None:
    collate_fn = engine.state.dataloader.collate_fn

    if isinstance(collate_fn, Collator):
        if isinstance(mode, ModeKeys):
            mode = mode.value

        getattr(collate_fn, mode)()
        engine.logger.info("Collator for [%s] switched to [%s] mode.", dataset, mode)


def build_evaluate_handler(
    evaluator: Engine,
    mode: str,
    model: nn.Module,
    dataflow: DataLoader,
    best_model_handler: Optional[Callable[[Engine], None]] = None,
):
    def evaluate_handler(_):
        # Ensure model is saved by rank 0 before possibly loading it.
        idist.barrier()

        evaluator.logger.info("Evaluate on [%s]", mode)
        if best_model_handler is not None:
            if best_model_handler.last_checkpoint is not None:
                checkpoint = os.path.join(
                    best_model_handler.save_handler.dirname,
                    best_model_handler.last_checkpoint,
                )
                evaluator.logger.info(
                    "Loading checkpoint %r before evaluate.",
                    checkpoint,
                )
                checkpoint = torch.load(checkpoint, map_location="cpu")
                Checkpoint.load_objects(to_load={"model": model}, checkpoint=checkpoint)

        evaluator.run(dataflow)

        evaluator.logger.info("Evaluate metrics [%s]", mode)
        for key, metric in sorted(evaluator.state.metrics.items(), key=lambda x: x[0]):
            # Ignote dict metrics flattened by ignite.
            if isinstance(metric, Mapping):
                continue

            evaluator.logger.info("%s = %s", key, metric)

    return evaluate_handler


def setup_progress_bar(
    params: Params, trainer: Engine, evaluators: Mapping[str, Engine]
) -> None:
    if params.pbar_steps > 0:
        ProgressBar(ncols=0).attach(
            trainer,
            metric_names="all",
            event_name=Events.ITERATION_COMPLETED(every=params.pbar_steps),
        )

        for evaluator in evaluators.values():
            ProgressBar(ncols=0).attach(
                evaluator,
                event_name=Events.ITERATION_COMPLETED(every=params.pbar_steps),
            )


def setup_lr_scheduler(
    params: Params, trainer: Engine, lr_scheduler: _LRScheduler
) -> None:
    if lr_scheduler is not None:
        if params.schedule_steps == "epoch" or params.schedule_steps > 0:
            trainer.add_event_handler(
                get_event(params.schedule_steps),
                lambda _: cast(_LRScheduler, lr_scheduler).step(),
            )
        else:
            raise ValueError(
                '`schedule_steps` must be positve or "epoch"'
                " when `lr_scheduler` is passed."
            )

    else:
        trainer.logger.warning("LR scheduler not set.")


def setup_save_best_model_handler(
    params: Params, trainer: Engine, evaluator: Engine, model: nn.Module
) -> Union[Checkpoint, None]:
    handler = None
    if params.save_best_models > 0:
        if params.eval_metric is None:
            raise ValueError("`eval_metric` must set when `save_best_models` is set.")

        handler = save_best_model_by_val_score(
            params.save_path,
            evaluator,
            model,
            params.eval_metric,
            n_saved=params.save_best_models,
            trainer=trainer,
        )
        trainer.logger.info(
            "Save best model hander set with `eval_metric` = %s"
            ", `save_best_models` = %d.",
            params.eval_metric,
            params.save_best_models,
        )
    else:
        trainer.logger.warning("Save best model handler not set.")

    return handler


def setup_early_stopping_handler(
    params: Params, trainer: Engine, evaluator: Engine
) -> Union[EarlyStopping, None]:
    handler = None
    if params.early_stopping:
        if params.eval_metric is None or params.patience is None:
            raise ValueError(
                "`eval_metric` and `patience` must set when `early_stopping` is set"
            )

        if params.save_best_models < 0:
            trainer.logger.warning(
                "Early stopping is set, but best model is not saved."
            )

        handler = add_early_stopping_by_val_score(
            params.patience,
            evaluator,
            trainer,
            params.eval_metric,
        )
        trainer.logger.info(
            "Early stopping is set with `eval_metric` = %s and `patience` = %d.",
            params.eval_metric,
            params.patience,
        )

    return handler


def setup_evaluate_handlers(
    params: Params,
    trainer: Engine,
    model: nn.Module,
    evaluators: Mapping[str, Engine],
    dataflows: Mapping[str, DataLoader],
) -> dict[str, Any]:
    handlers = dict.fromkeys(
        [
            "best_model_handler",
            "early_stopping_handler",
            "train_evaluate_handler",
            "val_evaluate_handler",
            "test_evaluate_handler",
        ]
    )

    if params.eval_steps == "epoch" or params.eval_steps > 0:
        for mode in ["train", "val"]:
            if mode == "train" and not params.eval_train:
                continue

            evaluate_handler = build_evaluate_handler(
                evaluators[mode], mode, model, dataflows[mode]
            )
            trainer.add_event_handler(get_event(params.eval_steps), evaluate_handler)
            handlers[f"{mode}_evaluate_handler"] = evaluate_handler

            trainer.logger.info("Setup evaluator for [%s]", mode)

        handlers["best_model_handler"] = setup_save_best_model_handler(
            params, trainer, evaluators["val"], model
        )

        handlers["early_stopping_handler"] = setup_early_stopping_handler(
            params, trainer, evaluators["val"]
        )
    else:
        trainer.logger.warning("Evaluate handlers not set")
        if params.early_stopping:
            raise ValueError(
                "Evaluate handlers must set when `early_stopping` is set"
                ", check `eval_steps`, `eval_metric` and `patience`."
            )

    test_evaluator = evaluators.get("test")
    if test_evaluator is not None:
        test_loader = dataflows.get("test")
        if test_loader is None:
            raise ValueError(
                "`test_loader` must not be None when `test_evaluator` is passed."
            )

        evaluate_handler = build_evaluate_handler(
            test_evaluator, "test", model, test_loader, handlers["best_model_handler"]
        )
        trainer.add_event_handler(Events.COMPLETED, evaluate_handler)

        if params.eval_before_training:
            trainer.add_event_handler(Events.STARTED, evaluate_handler)

        @trainer.on(Events.EXCEPTION_RAISED)
        def evaluate_test_on_exceptions(*_):
            evaluate_handler(test_evaluator)

        handlers["test_evaluate_handler"] = evaluate_handler

        trainer.logger.info("Setup evaluator for [test].")
    else:
        trainer.logger.warning("Test evaluate handler not set.")

    return handlers


def setup_logger_handlers(
    params: Params,
    trainer: Engine,
    model: nn.Module,
    optimizers: Optional[Union[Optimizer, Mapping[str, Optimizer]]] = None,
    evaluators: Optional[Mapping[str, Engine]] = None,
) -> dict[str, BaseLogger]:
    handlers = dict.fromkeys(["tensorboard_logger", "wandb_logger"])

    if params.log_steps < 0:
        trainer.logger.warning("Logger handlers not set.")

        return handlers

    if params.tensorboard:
        tensorboard_dir = os.path.join(params.save_path, "tensorboard")
        os.makedirs(tensorboard_dir, exist_ok=True)
        tensorboard_logger = setup_tb_logging(
            tensorboard_dir,
            trainer,
            optimizers=optimizers,
            evaluators=evaluators,
            log_steps=params.log_steps,
            model=model,
            include_weights_and_grads=params.debug,
        )
        handlers["tensorboard_logger"] = tensorboard_logger
        trainer.logger.info("Setup tensorboard logger: %s", tensorboard_dir)

    if params.wandb:
        # FIXME: Duplicated code with evaluator setup.
        # https://github.com/pytorch/ignite/issues/1476#issuecomment-826317167
        def filter_metrics(engine):
            engine.state.metrics = {
                k: v for k, v in engine.state.metrics.items() if not isinstance(v, dict)
            }

        if evaluators:
            for evaluator in evaluators.values():
                evaluator.add_event_handler(Events.COMPLETED, filter_metrics)

        wandb_dir = os.path.join(params.save_path, "wandb")
        os.makedirs(wandb_dir, exist_ok=True)
        wandb_logger = setup_wandb_logging(
            trainer,
            optimizers=optimizers,
            evaluators=evaluators,
            log_every_iters=params.log_steps,
            dir=wandb_dir,
            config=params.to_dict(),
            project=params.project,
            reinit=True,
        )
        wandb_logger.attach(trainer, lambda *_: wandb_logger.close, Events.COMPLETED)
        if params.debug:
            wandb_logger.watch(model, log="all", log_steps=params.log_steps)
        handlers["wandb_logger"] = wandb_logger
        trainer.logger.info("Setup wandb logger: %s", wandb_dir)

    return handlers


def setup_extra_handlers(
    params: Params,
    trainer: Engine,
    evaluators: Mapping[str, Engine],
    dataflows: Mapping[str, DataLoader],
    model: nn.Module,
    optimizer: Optimizer,
) -> dict[str, BaseLogger]:
    # Evaluation.
    setup_evaluate_handlers(params, trainer, model, evaluators, dataflows)

    # Loggers.
    loggers = {}
    if idist.get_rank() == 0:
        loggers = setup_logger_handlers(params, trainer, model, optimizer, evaluators)

    return loggers



# -*- coding: utf-8 -*-


class LinearDecayScheduledSampling(object):
    def __init__(
        self, max_epochs: int, k: float = 1.0, c: float = 1.0, eps: float = 0.0
    ):
        self.max_epochs = max_epochs
        self.k = k
        self.c = c
        self.eps = eps
        self.epoch = -1
        self.step()

    @property
    def teacher_forcing_ratio(self):
        return self.value

    def step(self) -> None:
        self.epoch += 1
        self.value = max(self.k - self.c * self.epoch / self.max_epochs, self.eps)



# -*- coding: utf-8 -*-

import torchlight.training.handlers
import torchlight.training.params
import torchlight.training.training
from torchlight.training.handlers import (
    setup_early_stopping_handler,
    setup_evaluate_handlers,
    setup_extra_handlers,
    setup_logger_handlers,
    setup_lr_scheduler,
    setup_progress_bar,
    setup_save_best_model_handler,
)
from torchlight.training.params import Params
from torchlight.training.training import (
    Dataflows,
    Metrics,
    create_engines,
    create_evaluator,
    create_evaluators,
    create_trainer,
    run,
    setup_engine,
    setup_env,
)

__all__ = [
    "setup_early_stopping_handler",
    "setup_evaluate_handlers",
    "setup_extra_handlers",
    "setup_logger_handlers",
    "setup_lr_scheduler",
    "setup_progress_bar",
    "setup_save_best_model_handler",
    "Params",
    "Dataflows" "Metrics",
    "create_engines",
    "create_evaluator",
    "create_evaluators",
    "create_trainer",
    "run",
    "setup_engine",
    "setup_env",
]



# -*- coding: utf-8 -*-
# pylint: disable=not-callable

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Optional, Union

from ignite.contrib.engines import common as logging_base
from ignite.contrib.handlers.tensorboard_logger import (
    GradsHistHandler,
    GradsScalarHandler,
    WeightsHistHandler,
    WeightsScalarHandler,
)
from ignite.engine import Events

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports
    import torch.nn as nn
    from ignite.contrib.handlers import TensorboardLogger
    from ignite.engine import Engine
    from torch.optim import Optimizer


def setup_tb_logging(
    log_dir: str,
    trainer: Engine,
    optimizers: Optional[Union[Optimizer, Mapping[str, Optimizer]]] = None,
    evaluators: Optional[Mapping[str, Engine]] = None,
    log_steps: int = 1,
    model: Optional[nn.Module] = None,
    include_weights_and_grads: bool = True,
    **kwargs,
) -> TensorboardLogger:
    tb_logger = logging_base.setup_tb_logging(
        log_dir,
        trainer,
        optimizers=optimizers,
        evaluators=evaluators,
        log_every_iters=log_steps,
        **kwargs,
    )

    if include_weights_and_grads:
        tb_logger.attach(
            trainer,
            log_handler=WeightsScalarHandler(model),
            event_name=Events.ITERATION_COMPLETED(every=log_steps),
        )

        tb_logger.attach(
            trainer,
            log_handler=WeightsHistHandler(model),
            event_name=Events.ITERATION_COMPLETED(every=log_steps),
        )

        tb_logger.attach(
            trainer,
            log_handler=GradsScalarHandler(model),
            event_name=Events.ITERATION_COMPLETED(every=log_steps),
        )

        tb_logger.attach(
            trainer,
            log_handler=GradsHistHandler(model),
            event_name=Events.ITERATION_COMPLETED(every=log_steps),
        )

    tb_logger.attach(trainer, lambda *args, **kwargs: tb_logger.close, Events.COMPLETED)

    return tb_logger



# -*- coding: utf-8 -*-

from __future__ import annotations

import logging
import os
import traceback
from collections.abc import Callable
from typing import Mapping, Optional, Union

import ignite.distributed as idist
import torch
import torch.nn as nn
from carton.logger import log_dict
from carton.logger import setup_logger as carton_setup_logger
from carton.random import set_seed
from ignite.contrib.engines.common import ProgressBar, setup_common_training_handlers
from ignite.contrib.handlers.base_logger import BaseLogger
from ignite.engine import Engine, Events
from ignite.handlers import DiskSaver
from ignite.metrics import BatchWise, EpochWise, Metric
from ignite.utils import convert_tensor, setup_logger
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader

from torchlight.training.handlers import handle_collator_mode, setup_extra_handlers
from torchlight.training.params import Params
from torchlight.utils import ModeKeys

logger = logging.getLogger(__name__)

Dataflows = Mapping[str, DataLoader]
Metrics = Mapping[str, Metric]


def setup_env(params: Params):
    set_seed(params["seed"])

    os.makedirs(os.path.dirname(params["log_file"]), exist_ok=True)
    carton_setup_logger(level=params["log_level"], filename=params["log_file"])

    params.to_yaml(os.path.join(params.save_path, "params.yaml"))


def setup_engine(
    engine: Engine,
    name: str,
    log_file: Optional[str] = None,
    metrics: Optional[Metrics] = None,
    train: bool = True,
) -> Engine:
    engine.logger = setup_logger(name, filepath=log_file)

    if metrics:
        for metric_name, metric in metrics.items():
            metric.attach(
                engine, metric_name, usage=BatchWise() if train else EpochWise()
            )
    return engine


def create_trainer(
    train_step: Callable,
    params: Params,
    model: Union[nn.Module, nn.parallel.DistributedDataParallel],
    criteria: nn.Module,
    optimizer: Optimizer,
    lr_scheduler: Optional[_LRScheduler] = None,
    metrics: Optional[Metrics] = None,
    name: str = "trainer",
    with_handlers: bool = True,
) -> Engine:
    def step(engine, batch):
        model.train()
        batch = convert_tensor(
            batch, device=idist.device(), non_blocking=params.non_blocking
        )
        output = train_step(
            engine,
            model.module
            if isinstance(model, nn.parallel.DistributedDataParallel)
            else model,
            batch,
            criteria,
        )
        loss = output["loss"]
        loss.backward()
        if (
            not params.gradient_accumulation_steps
            or engine.state.iteration % params.gradient_accumulation_steps == 0
        ):
            if params.max_grad_norm is not None:
                torch.nn.utils.clip_grad_norm_(model.parameters(), params.max_grad_norm)
            optimizer.step()
            optimizer.zero_grad()
        engine.state.metrics["loss"] = loss.item()

        return output

    trainer = Engine(step)
    trainer = setup_engine(
        trainer, name, log_file=params.log_file, metrics=metrics, train=True
    )

    if with_handlers:
        if params.pbar_steps > 0:
            ProgressBar(ncols=0).attach(
                trainer,
                metric_names="all",
                event_name=Events.ITERATION_COMPLETED(every=params.pbar_steps),
                closing_event_name=Events.EPOCH_COMPLETED,
            )

        trainer.add_event_handler(
            Events.EPOCH_STARTED,
            handle_collator_mode,
            "train",
            ModeKeys.TRAIN,
        )

        to_save = {
            "trainer": trainer,
            "model": model,
            "optimizer": optimizer,
        }
        if lr_scheduler is not None:
            to_save.update({"lr_scheduler": lr_scheduler})

        setup_common_training_handlers(
            trainer,
            train_sampler=None,  # TODO
            to_save=to_save,
            save_every_iters=params.save_steps,
            lr_scheduler=lr_scheduler,  # TODO
            with_gpu_stats=False,
            output_names=["loss"],
            with_pbars=False,
            with_pbar_on_iters=False,
            log_every_iters=params.pbar_steps,
            stop_on_nan=True,
            clear_cuda_cache=False,
            save_handler=DiskSaver(params.save_path, require_empty=False),
        )
        logger.info("Setup common training handlers.")

        @trainer.on(Events.EXCEPTION_RAISED)
        def handle_exceptions(engine, e):
            if isinstance(e, KeyboardInterrupt):
                engine.logger.info("KeyboardInterrupt caught. Exiting gracefully.")
                trainer.terminate()
            else:
                traceback.print_exc()
                raise e

        logger.info("Setup exception handler.")

    return trainer


def create_evaluator(
    eval_step: Callable,
    params: Params,
    model: Union[nn.Module, nn.parallel.DistributedDataParallel],
    tag: str,
    metrics: Optional[Metrics] = None,
    with_handlers: bool = True,
) -> Engine:
    @torch.no_grad()
    def step(engine, batch):
        model.eval()
        batch = convert_tensor(
            batch, device=idist.device(), non_blocking=params.non_blocking
        )
        output = eval_step(
            engine,
            model.module
            if isinstance(model, nn.parallel.DistributedDataParallel)
            else model,
            batch,
        )

        return output

    evaluator = Engine(step)
    evaluator = setup_engine(
        evaluator,
        f"evaluator/{tag}",
        log_file=params.log_file,
        metrics=metrics,
        train=False,
    )

    if with_handlers:
        ProgressBar(ncols=0, desc=f"Evaluation [{tag}]").attach(
            evaluator,
            event_name=Events.ITERATION_COMPLETED(every=params.pbar_steps),
            closing_event_name=Events.EPOCH_COMPLETED,
        )

        evaluator.add_event_handler(
            Events.EPOCH_STARTED,
            handle_collator_mode,
            tag,
            ModeKeys.EVAL,
        )

    return evaluator


def create_evaluators(
    eval_step: Callable,
    params: Params,
    model: nn.Module,
    metrics: Optional[Metrics] = None,
    with_handlers: bool = True,
) -> dict[str, Engine]:
    return {
        mode: create_evaluator(
            eval_step, params, model, mode, metrics=metrics, with_handlers=with_handlers
        )
        for mode in ["train", "val", "test"]
    }


def create_engines(
    params: Params,
    train_step: Callable,
    eval_step: Callable,
    dataflows: Dataflows,
    model: Union[nn.Module, nn.parallel.DistributedDataParallel],
    criteria: nn.Module,
    optimizer: Optimizer,
    lr_scheduler: Optional[_LRScheduler] = None,
    train_metrics: Optional[Metrics] = None,
    eval_metrics: Optional[Metrics] = None,
    metrics: Optional[Metrics] = None,
    with_handlers: bool = True,
) -> Union[
    tuple[Engine, dict[str, Engine]],
    tuple[Engine, dict[str, Engine], dict[str, BaseLogger]],
]:
    # pylint: disable=too-many-arguments
    logger.info("Creating engines with params:")
    log_dict(logger, params.to_dict())

    if metrics:
        if train_metrics or eval_metrics:
            raise ValueError(
                "Either `metrics` or `train_metrics` and `eval_metrics`"
                " should be given"
            )
        train_metrics = metrics
        eval_metrics = metrics

    trainer = create_trainer(
        train_step,
        params,
        model,
        criteria,
        optimizer,
        lr_scheduler=lr_scheduler,
        metrics=train_metrics,
        with_handlers=with_handlers,
    )
    evaluators = create_evaluators(
        eval_step, params, model, eval_metrics, with_handlers=with_handlers
    )

    if with_handlers:
        loggers = setup_extra_handlers(
            params, trainer, evaluators, dataflows, model, optimizer
        )

        return trainer, evaluators, loggers

    return trainer, evaluators


def run(fn: Callable, params: Params, *args, **kwargs) -> None:
    with idist.Parallel(
        backend=params.backend,
        nnodes=params.nnodes,
        node_rank=params.node_rank,
        nproc_per_node=params.nproc_per_node,
        master_addr=params.master_addr,
        master_port=params.master_port,
        init_method=params.init_method,
    ) as parallel:
        parallel.run(fn, params, *args, **kwargs)



# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F


class ScaledDotProductAttention(nn.Module):
    def __init__(self, hidden_size, dropout=0.0):
        super().__init__()
        self.hidden_size = hidden_size
        self.scaler = torch.sqrt(torch.tensor(self.hidden_size).float())
        self.dropout = nn.Dropout(dropout)

    def forward(self, query, key, value, mask=None, return_attns=False):
        # `query`, `key`, `value`
        # shape: batch_size x max_len x hidden_size
        assert query.dim() == 3
        assert key.dim() == 3
        assert value.dim() == 3

        if query.size()[-1] != key.size()[-1]:
            raise ValueError(
                "mismatch size (dim: -1): `query`: {} and `key`: {}".format(
                    list(query.size()), list(key.size())
                )
            )

        if query.size()[-1] != self.hidden_size:
            raise ValueError(
                (
                    "mismatch size (dim: -1): `query` or `key` [{}]"
                    " and `hidden_size` [{}]"
                ).format(query.size()[-1], self.hidden_size)
            )

        # Compute raw attention scores, i.e. A = QK^{T}.
        # shape: batch_size x max_len x max_len
        attns = torch.bmm(query, key.transpose(1, 2))
        attns /= self.scaler

        # Normalization.
        if mask is not None:
            if mask.size() != attns.size():
                raise ValueError(
                    "mismatch size: `attns`: {} and `mask`: {}".format(
                        list(attns.size()), list(mask.size())
                    )
                )
            attns.masked_fill_(mask, float("-inf"))

        if return_attns:
            return value, attns

        # Apply softmax and dropout.
        attns = F.softmax(attns, dim=-1)
        attns = self.dropout(attns)

        # Weight `value` by attentions.
        outputs = torch.bmm(attns, value)

        return outputs


class FeedForwardAttention(nn.Module):
    def __init__(self, in_channels, out_channels, activation=None, dropout=0.0):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        if activation is None:
            activation = nn.LeakyReLU()
        self.activation = activation
        self.dropout = nn.Dropout(dropout)

        self.W_query = nn.Linear(in_channels, out_channels)
        self.W_key = nn.Linear(in_channels, out_channels)
        self.W_value = nn.Linear(in_channels, out_channels)
        self.fc_query = nn.Linear(out_channels, 1)
        self.fc_key = nn.Linear(out_channels, 1)

    def forward(self, query, key, value, mask=None, return_attns=False):
        assert query.dim() == 3
        assert key.dim() == 3
        assert value.dim() == 3

        if len(set([query.size()[-1], key.size()[-1], value.size()[-1]])) != 1:
            raise ValueError(
                (
                    "mismatch size (dim: -1) of `query`: {} and `key`: {}"
                    " and `value`: {}"
                ).format(list(query.size()), list(key.size()), list(value.size()))
            )

        # `query`, `key`, `value` should have same hidden size.
        if query.size()[-1] != self.in_channels:
            raise ValueError(
                (
                    "mismatch size: dim -1 of `query` or `key` or `value` [{}]"
                    " and `in_channels` [{}]"
                ).format(query.size()[-1], self.in_channels)
            )

        # `query`, `key`, `value`
        # shape: batch_size x max_len x in_channels
        # `weighted_query`, `weighted_key`, `weighted_value`
        # shape: batch_size x max_len x out_channels
        w_key = self.W_key(key)
        w_query = self.W_query(query) if query is not key else w_key
        w_value = self.W_value(value) if value is not key else w_key

        # Feed to a single layer MLP.
        # `attns`
        # shape: batch_size x max_len x max_len
        fc_query = self.fc_query(w_query)
        fc_key = self.fc_key(w_key)
        attns = fc_query + fc_key.transpose(1, 2)

        # Normalization.
        if mask is not None:
            if mask.size() != attns.size():
                raise ValueError(
                    "mismatch size: `attns`: {} and `mask`: {}".format(
                        list(attns.size()), list(mask.size())
                    )
                )
            attns.masked_fill_(mask, float("-inf"))

        if return_attns:
            return w_value, attns

        # Apply softmax and dropout.
        attns = F.softmax(attns, dim=-1)
        attns = self.dropout(attns)

        # Weight `w_value` by attentions.
        outputs = torch.bmm(attns, w_value)

        # Apply activation.
        attns = self.activation(attns)

        return outputs


class BidirectionalAttention(nn.Module):
    def forward(self, v1, v1_mask, v2, v2_mask):
        similarity_matrix = v1.bmm(v2.transpose(2, 1))

        v2_v1_attn = F.softmax(
            similarity_matrix.masked_fill(v1_mask.unsqueeze(2), -float("inf")), dim=1
        )
        v1_v2_attn = F.softmax(
            similarity_matrix.masked_fill(v2_mask.unsqueeze(1), -float("inf")), dim=2
        )

        attended_v1 = v1_v2_attn.bmm(v2)
        attended_v2 = v2_v1_attn.transpose(1, 2).bmm(v1)

        attended_v1.masked_fill_(v1_mask.unsqueeze(2), 0)
        attended_v2.masked_fill_(v2_mask.unsqueeze(2), 0)

        return attended_v1, attended_v2



# -*- coding: utf-8 -*-

from __future__ import annotations

import argparse
import copy
import datetime
import os
from collections.abc import ItemsView, KeysView, ValuesView
from typing import Optional, Type, TypeVar

from carton.params import Params as ParamDict
from carton.random import random_state

from torchlight.utils import device


class Params(object):
    # pylint: disable=too-many-instance-attributes

    T = TypeVar("T", bound="Params")

    def __init__(self, **kwargs):
        # Project
        self.project = kwargs.get("project")
        if not self.project:
            raise ValueError("`project` must not be empty")

        # Datasets
        self.data_dir = kwargs.get("data_dir")
        self.train_data = kwargs.get("train_data")
        self.val_data = kwargs.get("val_data")
        self.test_data = kwargs.get("test_data")
        self.train_size = kwargs.get("train_size")
        self.val_size = kwargs.get("val_size", 0.1)
        self.test_size = kwargs.get("test_size", 0.2)
        self.max_length = kwargs.get("max_length")

        # Training
        self.backend = kwargs.get("backend")
        self.nproc_per_node = kwargs.get("nproc_per_node")
        self.nnodes = kwargs.get("nnodes")
        self.node_rank = kwargs.get("node_rank")
        self.master_addr = kwargs.get("master_addr")
        self.master_port = kwargs.get("master_port")
        self.init_method = kwargs.get("init_method")
        self.seed = random_state(kwargs.get("seed"))
        save_path = os.getenv("SAVE_PATH")
        if not save_path:
            save_path = kwargs.get("save_path", "output/")
        self.save_path = os.path.join(
            save_path, self.project, datetime.datetime.now().isoformat()
        )
        self.log_file = os.path.join(self.save_path, "output.log")
        self.max_epochs = kwargs.get("max_epochs", 10)
        self.lr = kwargs.get("lr")
        if self.lr is None:
            raise ValueError("`lr` must not be None")
        self.lr_warmup = kwargs.get("lr_warmup", 0.1)
        self.weight_decay = kwargs.get("weight_decay", 0.0)
        self.max_grad_norm = kwargs.get("max_grad_norm")
        self.schedule_steps = kwargs.get("schedule_steps", -1)
        self.gradient_accumulation_steps = kwargs.get("gradient_accumulation_steps")

        # Dataloader
        self.batch_size = kwargs.get("batch_size", 32)
        self.train_batch_size = kwargs.get("train_batch_size", self.batch_size)
        self.eval_batch_size = kwargs.get("eval_batch_size", self.batch_size)
        self.predict_batch_size = kwargs.get("predict_batch_size", self.batch_size)
        self.device = device(kwargs.get("device", "cuda"))
        self.pin_memory = kwargs.get("pin_memory", True)
        self.non_blocking = kwargs.get("non_blocking", True)
        self.num_workers = kwargs.get("num_workers", 0)

        # Checkpoint
        self.save_steps = kwargs.get("save_steps", 1000)

        # Evaluation
        self.eval_train = kwargs.get("eval_train", False)
        self.eval_steps = kwargs.get("eval_steps", -1)
        self.eval_metric = kwargs.get("eval_metric")
        self.save_best_models = kwargs.get(
            "save_best_models",
            1
            if (self.eval_steps == "epoch" or self.eval_steps > 0)
            and self.eval_metric is not None
            else 0,
        )
        self.eval_before_training = kwargs.get("eval_before_training", False)

        # Early Stopping
        self.early_stopping = kwargs.get("early_stopping", False)
        self.patience = kwargs.get("patience")

        # Logging
        self.pbar_steps = kwargs.get("pbar_steps", 1)
        self.log_steps = kwargs.get("log_steps", -1)
        self.log_level = kwargs.get("log_level", "INFO")
        self.tensorboard = kwargs.get("tensorboard", False)
        self.wandb = kwargs.get("wandb", False)
        self.debug = kwargs.get("debug", False)
        self.sample_dir = kwargs.get(
            "sample_dir", os.path.join(self.save_path, "samples")
        )

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def to_dict(self) -> dict:
        return copy.deepcopy(self.__dict__)

    def keys(self) -> KeysView:
        return self.__dict__.keys()

    def values(self) -> ValuesView:
        return self.__dict__.values()

    def items(self) -> ItemsView:
        return self.__dict__.items()

    def to_yaml(self, filename: Optional[str] = None) -> None:
        if not filename:
            filename = os.path.join(self.save_path, "params.yaml")

        ParamDict(**self.to_dict()).to_yaml(filename)

    def to_json(self, filename: Optional[str] = None) -> None:
        if not filename:
            filename = os.path.join(self.save_path, "params.json")

        ParamDict(**self.to_dict()).to_json(filename)

    def to_toml(self, filename: Optional[str] = None) -> None:
        if not filename:
            filename = os.path.join(self.save_path, "params.toml")

        ParamDict(**self.to_dict()).to_toml(filename)

    @classmethod
    def from_yaml(cls: Type[T], filename: str) -> T:
        return cls(**ParamDict.from_yaml(filename))

    @classmethod
    def from_json(cls: Type[T], filename: str) -> T:
        return cls(**ParamDict.from_json(filename))

    @classmethod
    def from_toml(cls: Type[T], filename: str) -> T:
        return cls(**ParamDict.from_toml(filename))


def main():
    # pylint: disable=import-outside-toplevel

    def parse_args():
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
        parser.add_argument(
            "--output",
            type=str,
            default="template.yaml",
            help="Output template file name.",
        )

        return parser.parse_args()  # pylint: disable=redefined-outer-name

    args = parse_args()
    params = ParamDict(**Params(project="NOT SET", lr="NOT SET").to_dict())
    _, ext = os.path.splitext(args.output)
    exts = {".yaml", ".json", ".toml"}
    if ext not in exts:
        raise RuntimeError(
            (f"Unsupported file type: {ext}, supported file types are: {exts}")
        )

    getattr(params, f"to_{ext[1:]}")(args.output)


if __name__ == "__main__":
    main()
